# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for bzr.

These check that it behaves properly when it's invoked through the regular
command-line interface. This doesn't actually run a new interpreter but
rather starts again from the run_brz function.
"""

from breezy import config, tests
from breezy.urlutils import joinpath

from ..test_bedding import override_whoami


class TestAnnotate(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        wt = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [("hello.txt", b"my helicopter\n"), ("nomail.txt", b"nomail\n")]
        )
        wt.add(["hello.txt"])
        self.revision_id_1 = wt.commit(
            "add hello", committer="test@user", timestamp=1165960000.00, timezone=0
        )
        wt.add(["nomail.txt"])
        self.revision_id_2 = wt.commit(
            "add nomail", committer="no mail", timestamp=1165970000.00, timezone=0
        )
        self.build_tree_contents([("hello.txt", b"my helicopter\nyour helicopter\n")])
        self.revision_id_3 = wt.commit(
            "mod hello", committer="user@test", timestamp=1166040000.00, timezone=0
        )
        self.build_tree_contents(
            [
                (
                    "hello.txt",
                    b"my helicopter\nyour helicopter\nall of\nour helicopters\n",
                )
            ]
        )
        self.revision_id_4 = wt.commit(
            "mod hello", committer="user@test", timestamp=1166050000.00, timezone=0
        )

    def test_help_annotate(self):
        """Annotate command exists."""
        out, err = self.run_bzr("--no-plugins annotate --help")

    def test_annotate_cmd(self):
        out, err = self.run_bzr("annotate hello.txt")
        self.assertEqual("", err)
        self.assertEqualDiff(
            """\
1   test@us | my helicopter
3   user@te | your helicopter
4   user@te | all of
            | our helicopters
""",
            out,
        )

    def test_annotate_cmd_full(self):
        out, err = self.run_bzr("annotate hello.txt --all")
        self.assertEqual("", err)
        self.assertEqualDiff(
            """\
1   test@us | my helicopter
3   user@te | your helicopter
4   user@te | all of
4   user@te | our helicopters
""",
            out,
        )

    def test_annotate_cmd_long(self):
        out, err = self.run_bzr("annotate hello.txt --long")
        self.assertEqual("", err)
        self.assertEqualDiff(
            """\
1   test@user 20061212 | my helicopter
3   user@test 20061213 | your helicopter
4   user@test 20061213 | all of
                       | our helicopters
""",
            out,
        )

    def test_annotate_cmd_show_ids(self):
        out, err = self.run_bzr("annotate hello.txt --show-ids")
        max_len = max(
            [len(self.revision_id_1), len(self.revision_id_3), len(self.revision_id_4)]
        )
        self.assertEqual("", err)
        self.assertEqualDiff(
            """\
%*s | my helicopter
%*s | your helicopter
%*s | all of
%*s | our helicopters
"""
            % (
                max_len,
                self.revision_id_1.decode("utf-8"),
                max_len,
                self.revision_id_3.decode("utf-8"),
                max_len,
                self.revision_id_4.decode("utf-8"),
                max_len,
                "",
            ),
            out,
        )

    def test_no_mail(self):
        out, err = self.run_bzr("annotate nomail.txt")
        self.assertEqual("", err)
        self.assertEqualDiff(
            """\
2   no mail | nomail
""",
            out,
        )

    def test_annotate_cmd_revision(self):
        out, err = self.run_bzr("annotate hello.txt -r1")
        self.assertEqual("", err)
        self.assertEqualDiff(
            """\
1   test@us | my helicopter
""",
            out,
        )

    def test_annotate_cmd_revision3(self):
        out, err = self.run_bzr("annotate hello.txt -r3")
        self.assertEqual("", err)
        self.assertEqualDiff(
            """\
1   test@us | my helicopter
3   user@te | your helicopter
""",
            out,
        )

    def test_annotate_cmd_unknown_revision(self):
        out, err = self.run_bzr("annotate hello.txt -r 10", retcode=3)
        self.assertEqual("", out)
        self.assertContainsRe(err, "Requested revision: '10' does not exist")

    def test_annotate_cmd_two_revisions(self):
        out, err = self.run_bzr("annotate hello.txt -r1..2", retcode=3)
        self.assertEqual("", out)
        self.assertEqual(
            "brz: ERROR: brz annotate --revision takes"
            " exactly one revision identifier\n",
            err,
        )


class TestSimpleAnnotate(tests.TestCaseWithTransport):
    """Annotate tests with no complex setup."""

    def _setup_edited_file(self, relpath="."):
        """Create a tree with a locally edited file."""
        tree = self.make_branch_and_tree(relpath)
        file_relpath = joinpath(relpath, "file")
        self.build_tree_contents([(file_relpath, b"foo\ngam\n")])
        tree.add("file")
        tree.commit("add file", committer="test@host", rev_id=b"rev1")
        self.build_tree_contents([(file_relpath, b"foo\nbar\ngam\n")])
        return tree

    def test_annotate_cmd_revspec_branch(self):
        tree = self._setup_edited_file("trunk")
        tree.branch.create_checkout(self.get_url("work"), lightweight=True)
        out, err = self.run_bzr(
            ["annotate", "file", "-r", "branch:../trunk"], working_dir="work"
        )
        self.assertEqual("", err)
        self.assertEqual("1   test@ho | foo\n            | gam\n", out)

    def test_annotate_edited_file(self):
        self._setup_edited_file()
        self.overrideEnv("BRZ_EMAIL", "current@host2")
        out, err = self.run_bzr("annotate file")
        self.assertEqual(
            "1   test@ho | foo\n2?  current | bar\n1   test@ho | gam\n", out
        )

    def test_annotate_edited_file_no_default(self):
        # Ensure that when no username is available annotate still works.
        override_whoami(self)
        self._setup_edited_file()
        out, err = self.run_bzr("annotate file")
        self.assertEqual(
            "1   test@ho | foo\n2?  local u | bar\n1   test@ho | gam\n", out
        )

    def test_annotate_edited_file_show_ids(self):
        self._setup_edited_file()
        self.overrideEnv("BRZ_EMAIL", "current@host2")
        out, err = self.run_bzr("annotate file --show-ids")
        self.assertEqual("    rev1 | foo\ncurrent: | bar\n    rev1 | gam\n", out)

    def _create_merged_file(self):
        """Create a file with a pending merge and local edit."""
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("file", b"foo\ngam\n")])
        tree.add("file")
        tree.commit("add file", rev_id=b"rev1", committer="test@host")
        # right side
        self.build_tree_contents([("file", b"foo\nbar\ngam\n")])
        tree.commit("right", rev_id=b"rev1.1.1", committer="test@host")
        tree.pull(tree.branch, True, b"rev1")
        # left side
        self.build_tree_contents([("file", b"foo\nbaz\ngam\n")])
        tree.commit("left", rev_id=b"rev2", committer="test@host")
        # merge
        tree.merge_from_branch(tree.branch, b"rev1.1.1")
        # edit the file to be 'resolved' and have a further local edit
        self.build_tree_contents([("file", b"local\nfoo\nbar\nbaz\ngam\n")])
        return tree

    def test_annotated_edited_merged_file_revnos(self):
        wt = self._create_merged_file()
        out, err = self.run_bzr(["annotate", "file"])
        email = config.extract_email_address(wt.branch.get_config_stack().get("email"))
        self.assertEqual(
            "3?    %-7s | local\n"
            "1     test@ho | foo\n"
            "1.1.1 test@ho | bar\n"
            "2     test@ho | baz\n"
            "1     test@ho | gam\n" % email[:7],
            out,
        )

    def test_annotated_edited_merged_file_ids(self):
        self._create_merged_file()
        out, err = self.run_bzr(["annotate", "file", "--show-ids"])
        self.assertEqual(
            "current: | local\n"
            "    rev1 | foo\n"
            "rev1.1.1 | bar\n"
            "    rev2 | baz\n"
            "    rev1 | gam\n",
            out,
        )

    def test_annotate_empty_file(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("empty", b"")])
        tree.add("empty")
        tree.commit("add empty file")
        out, err = self.run_bzr(["annotate", "empty"])
        self.assertEqual("", out)

    def test_annotate_removed_file(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("empty", b"")])
        tree.add("empty")
        tree.commit("add empty file")
        # delete the file.
        tree.remove("empty")
        tree.commit("remove empty file")
        out, err = self.run_bzr(["annotate", "-r1", "empty"])
        self.assertEqual("", out)

    def test_annotate_empty_file_show_ids(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("empty", b"")])
        tree.add("empty")
        tree.commit("add empty file")
        out, err = self.run_bzr(["annotate", "--show-ids", "empty"])
        self.assertEqual("", out)

    def test_annotate_nonexistant_file(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["file"])
        tree.add(["file"])
        tree.commit("add a file")
        out, err = self.run_bzr(["annotate", "doesnotexist"], retcode=3)
        self.assertEqual("", out)
        self.assertEqual("brz: ERROR: doesnotexist is not versioned.\n", err)

    def test_annotate_without_workingtree(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("empty", b"")])
        tree.add("empty")
        tree.commit("add empty file")
        bzrdir = tree.branch.controldir
        bzrdir.destroy_workingtree()
        self.assertFalse(bzrdir.has_workingtree())
        out, err = self.run_bzr(["annotate", "empty"])
        self.assertEqual("", out)

    def test_annotate_directory(self):
        """Test --directory option."""
        wt = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/hello.txt", b"my helicopter\n")])
        wt.add(["hello.txt"])
        wt.commit("commit", committer="test@user")
        out, err = self.run_bzr(["annotate", "-d", "a", "hello.txt"])
        self.assertEqualDiff("1   test@us | my helicopter\n", out)



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for brz missing."""

from breezy import osutils, tests


class TestMissing(tests.TestCaseWithTransport):
    def assertMessages(self, out, must_have=(), must_not_have=()):
        """Check if commit messages are in or not in the output."""
        for m in must_have:
            self.assertContainsRe(out, r"\nmessage:\n  {}\n".format(m))
        for m in must_not_have:
            self.assertNotContainsRe(out, r"\nmessage:\n  {}\n".format(m))

    def test_missing_quiet(self):
        # <https://bugs.launchpad.net/bzr/+bug/284748>
        # create a source branch
        #
        # XXX: This still needs a test that missing is quiet when there are
        # missing revisions.
        a_tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", b"initial\n")])
        a_tree.add("a")
        a_tree.commit(message="initial")

        out, err = self.run_bzr("missing -q .")
        self.assertEqual("", out)
        self.assertEqual("", err)

    def test_missing(self):
        missing_one = "You are missing 1 revision:"
        extra_one = "You have 1 extra revision:"

        # create a source branch
        a_tree = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/a", b"initial\n")])
        a_tree.add("a")
        a_tree.commit(message="initial")

        # clone and add a differing revision
        b_tree = a_tree.controldir.sprout("b").open_workingtree()
        self.build_tree_contents([("b/a", b"initial\nmore\n")])
        b_tree.commit(message="more")

        def run_missing(args, retcode=1, working_dir=None):
            out, err = self.run_bzr(
                ["missing"] + args, retcode=retcode, working_dir=working_dir
            )
            # we do not expect any error output.
            self.assertEqual("", err)
            return out.splitlines()

        def run_missing_a(args, retcode=1):
            return run_missing(["../a"] + args, retcode=retcode, working_dir="b")

        def run_missing_b(args, retcode=1):
            return run_missing(["../b"] + args, retcode=retcode, working_dir="a")

        # run missing in a against b
        # this should not require missing to take out a write lock on a
        # or b. So we take a write lock on both to test that at the same
        # time. This may let the test pass while the default branch is an
        # os-locking branch, but it will trigger failures with lockdir based
        # branches.
        a_branch = a_tree.branch
        a_branch.lock_write()
        b_branch = b_tree.branch
        b_branch.lock_write()

        lines = run_missing_b([])
        # we're missing the extra revision here
        self.assertEqual(missing_one, lines[0])
        # and we expect 8 lines of output which we trust at the moment to be
        # good.
        self.assertEqual(8, len(lines))
        # unlock the branches for the rest of the test
        a_branch.unlock()
        b_branch.unlock()

        # get extra revision from b
        a_tree.merge_from_branch(b_branch)
        a_tree.commit(message="merge")

        # compare again, but now we have the 'merge' commit extra
        lines = run_missing_b([])
        self.assertEqual(extra_one, lines[0])
        self.assertLength(8, lines)

        lines2 = run_missing_b(["--mine-only"])
        self.assertEqual(lines, lines2)

        lines3 = run_missing_b(["--theirs-only"], retcode=0)
        self.assertEqualDiff("Other branch has no new revisions.", lines3[0])

        # relative to a, missing the 'merge' commit
        lines = run_missing_a([])
        self.assertEqual(missing_one, lines[0])
        self.assertLength(8, lines)

        lines2 = run_missing_a(["--theirs-only"])
        self.assertEqual(lines, lines2)

        lines3 = run_missing_a(["--mine-only"], retcode=0)
        self.assertEqualDiff("This branch has no new revisions.", lines3[0])

        lines4 = run_missing_a(["--short"])
        self.assertLength(4, lines4)

        lines4a = run_missing_a(["-S"])
        self.assertEqual(lines4, lines4a)

        lines5 = run_missing_a(["--line"])
        self.assertLength(2, lines5)

        lines6 = run_missing_a(["--reverse"])
        self.assertEqual(lines6, lines)

        lines7 = run_missing_a(["--show-ids"])
        self.assertLength(11, lines7)

        lines8 = run_missing_a(["--verbose"])
        self.assertEqual("modified:", lines8[-2])
        self.assertEqual("  a", lines8[-1])

        self.assertEqualDiff(
            "Other branch has no new revisions.",
            run_missing_b(["--theirs-only"], retcode=0)[0],
        )

        # after a pull we're back on track
        b_tree.pull(a_branch)
        self.assertEqualDiff(
            "Branches are up to date.", run_missing_b([], retcode=0)[0]
        )
        self.assertEqualDiff(
            "Branches are up to date.", run_missing_a([], retcode=0)[0]
        )
        # If you supply mine or theirs you only know one side is up to date
        self.assertEqualDiff(
            "This branch has no new revisions.",
            run_missing_a(["--mine-only"], retcode=0)[0],
        )
        self.assertEqualDiff(
            "Other branch has no new revisions.",
            run_missing_a(["--theirs-only"], retcode=0)[0],
        )

    def test_missing_filtered(self):
        # create a source branch
        a_tree = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/a", b"initial\n")])
        a_tree.add("a")
        a_tree.commit(message="r1")
        # clone and add differing revisions
        b_tree = a_tree.controldir.sprout("b").open_workingtree()

        for i in range(2, 6):
            a_tree.commit(message="a%d" % i)
            b_tree.commit(message="b%d" % i)

        # local
        out, err = self.run_bzr(
            "missing ../b --my-revision 3", retcode=1, working_dir="a"
        )
        self.assertMessages(out, ("a3", "b2", "b3", "b4", "b5"), ("a2", "a4"))

        out, err = self.run_bzr(
            "missing ../b --my-revision 3..4", retcode=1, working_dir="a"
        )
        self.assertMessages(out, ("a3", "a4"), ("a2", "a5"))

        # remote
        out, err = self.run_bzr("missing ../b -r 3", retcode=1, working_dir="a")
        self.assertMessages(out, ("a2", "a3", "a4", "a5", "b3"), ("b2", "b4"))

        out, err = self.run_bzr("missing ../b -r 3..4", retcode=1, working_dir="a")
        self.assertMessages(out, ("b3", "b4"), ("b2", "b5"))

        # both
        out, err = self.run_bzr(
            "missing ../b --my-revision 3..4 -r 3..4", retcode=1, working_dir="a"
        )
        self.assertMessages(out, ("a3", "a4", "b3", "b4"), ("a2", "a5", "b2", "b5"))

    def test_missing_check_last_location(self):
        # check that last location shown as filepath not file URL

        # create a source branch
        wt = self.make_branch_and_tree("a")
        b = wt.branch
        self.build_tree(["a/foo"])
        wt.add("foo")
        wt.commit("initial")

        location = osutils.getcwd() + "/a/"

        # clone
        b.controldir.sprout("b")

        # check last location
        lines, err = self.run_bzr("missing", working_dir="b")
        self.assertEqual(
            "Using saved parent location: {}\nBranches are up to date.\n".format(
                location
            ),
            lines,
        )
        self.assertEqual("", err)

    def test_missing_directory(self):
        """Test --directory option."""
        # create a source branch
        a_tree = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/a", b"initial\n")])
        a_tree.add("a")
        a_tree.commit(message="initial")

        # clone and add a differing revision
        b_tree = a_tree.controldir.sprout("b").open_workingtree()
        self.build_tree_contents([("b/a", b"initial\nmore\n")])
        b_tree.commit(message="more")

        out2, err2 = self.run_bzr("missing --directory a b", retcode=1)
        out1, err1 = self.run_bzr("missing ../b", retcode=1, working_dir="a")
        self.assertEqualDiff(out1, out2)
        self.assertEqualDiff(err1, err2)

    def test_missing_tags(self):
        """Test showing tags."""
        # create a source branch
        a_tree = self.make_branch_and_tree("a")
        self.build_tree_contents([("a/a", b"initial\n")])
        a_tree.add("a")
        a_tree.commit(message="initial")

        # clone and add a differing revision
        b_tree = a_tree.controldir.sprout("b").open_workingtree()
        self.build_tree_contents([("b/a", b"initial\nmore\n")])
        b_tree.commit(message="more")
        b_tree.branch.tags.set_tag("a-tag", b_tree.last_revision())

        for log_format in ["long", "short", "line"]:
            out, err = self.run_bzr(
                f"missing --log-format={log_format} ../a", working_dir="b", retcode=1
            )
            self.assertContainsString(out, "a-tag")

            out, err = self.run_bzr(
                f"missing --log-format={log_format} ../b", working_dir="a", retcode=1
            )
            self.assertContainsString(out, "a-tag")



# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from ..workspace import Workspace, WorkspaceDirty, check_clean_tree
from . import TestCaseWithTransport, features, multiply_scenarios
from .scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class CheckCleanTreeTests(TestCaseWithTransport):
    def make_test_tree(self, format=None):
        tree = self.make_branch_and_tree(".", format=format)
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/control",
                    """\
Source: blah
Vcs-Git: https://example.com/blah
Testsuite: autopkgtest

Binary: blah
Arch: all

""",
                ),
                ("debian/changelog", "Some contents"),
            ]
        )
        tree.add(["debian", "debian/changelog", "debian/control"])
        tree.commit("Initial thingy.")
        return tree

    def test_pending_changes(self):
        tree = self.make_test_tree()
        self.build_tree_contents([("debian/changelog", "blah")])
        with tree.lock_write():
            self.assertRaises(WorkspaceDirty, check_clean_tree, tree)

    def test_pending_changes_bzr_empty_dir(self):
        # See https://bugs.debian.org/914038
        tree = self.make_test_tree(format="bzr")
        self.build_tree_contents([("debian/upstream/",)])
        with tree.lock_write():
            self.assertRaises(WorkspaceDirty, check_clean_tree, tree)

    def test_pending_changes_git_empty_dir(self):
        # See https://bugs.debian.org/914038
        tree = self.make_test_tree(format="git")
        self.build_tree_contents([("debian/upstream/",)])
        with tree.lock_write():
            check_clean_tree(tree)

    def test_pending_changes_git_dir_with_ignored(self):
        # See https://bugs.debian.org/914038
        tree = self.make_test_tree(format="git")
        self.build_tree_contents(
            [
                ("debian/upstream/",),
                ("debian/upstream/blah", ""),
                (".gitignore", "blah\n"),
            ]
        )
        tree.add(".gitignore")
        tree.commit("add gitignore")
        with tree.lock_write():
            check_clean_tree(tree)

    def test_extra(self):
        tree = self.make_test_tree()
        self.build_tree_contents([("debian/foo", "blah")])
        with tree.lock_write():
            self.assertRaises(WorkspaceDirty, check_clean_tree, tree)

    def test_subpath(self):
        tree = self.make_test_tree()
        self.build_tree_contents([("debian/foo", "blah"), ("foo/",)])
        tree.add("foo")
        tree.commit("add foo")
        with tree.lock_write():
            check_clean_tree(tree, tree.basis_tree(), subpath="foo")
            self.assertRaises(
                WorkspaceDirty, check_clean_tree, tree, tree.basis_tree(), subpath=""
            )

    def test_subpath_changed(self):
        tree = self.make_test_tree()
        self.build_tree_contents([("foo/",)])
        tree.add("foo")
        tree.commit("add foo")
        self.build_tree_contents([("debian/control", "blah")])
        with tree.lock_write():
            check_clean_tree(tree, tree.basis_tree(), subpath="foo")
            self.assertRaises(
                WorkspaceDirty, check_clean_tree, tree, tree.basis_tree(), subpath=""
            )


def vary_by_inotify():
    return [
        ("with_inotify", {"_use_inotify": True}),
        ("without_inotify", {"_use_inotify": False}),
    ]


def vary_by_format():
    return [
        ("bzr", {"_format": "bzr"}),
        ("git", {"_format": "git"}),
    ]


class WorkspaceTests(TestCaseWithTransport):
    scenarios = multiply_scenarios(
        vary_by_inotify(),
        vary_by_format(),
    )

    def setUp(self):
        super().setUp()
        if self._use_inotify:
            self.requireFeature(features.pyinotify)

    def test_root_add(self):
        tree = self.make_branch_and_tree(".", format=self._format)
        with Workspace(tree, use_inotify=self._use_inotify) as ws:
            self.build_tree_contents([("afile", "somecontents")])
            changes = [c for c in ws.iter_changes() if c.path[1] != ""]
            self.assertEqual(1, len(changes), changes)
            self.assertEqual((None, "afile"), changes[0].path)
            ws.commit(message="Commit message")
            self.assertEqual(list(ws.iter_changes()), [])
            self.build_tree_contents([("afile", "newcontents")])
            [change] = list(ws.iter_changes())
            self.assertEqual(("afile", "afile"), change.path)

    def test_root_remove(self):
        tree = self.make_branch_and_tree(".", format=self._format)
        self.build_tree_contents([("afile", "somecontents")])
        tree.add(["afile"])
        tree.commit("Afile")
        with Workspace(tree, use_inotify=self._use_inotify) as ws:
            os.remove("afile")
            changes = list(ws.iter_changes())
            self.assertEqual(1, len(changes), changes)
            self.assertEqual(("afile", None), changes[0].path)
            ws.commit(message="Commit message")
            self.assertEqual(list(ws.iter_changes()), [])

    def test_subpath_add(self):
        tree = self.make_branch_and_tree(".", format=self._format)
        self.build_tree(["subpath/"])
        tree.add("subpath")
        tree.commit("add subpath")
        with Workspace(tree, subpath="subpath", use_inotify=self._use_inotify) as ws:
            self.build_tree_contents([("outside", "somecontents")])
            self.build_tree_contents([("subpath/afile", "somecontents")])
            changes = [c for c in ws.iter_changes() if c.path[1] != "subpath"]
            self.assertEqual(1, len(changes), changes)
            self.assertEqual((None, "subpath/afile"), changes[0].path)
            ws.commit(message="Commit message")
            self.assertEqual(list(ws.iter_changes()), [])

    def test_dirty(self):
        tree = self.make_branch_and_tree(".", format=self._format)
        self.build_tree(["subpath"])
        self.assertRaises(
            WorkspaceDirty, Workspace(tree, use_inotify=self._use_inotify).__enter__
        )

    def test_reset(self):
        tree = self.make_branch_and_tree(".", format=self._format)
        with Workspace(tree, use_inotify=self._use_inotify) as ws:
            self.build_tree(["blah"])
            ws.reset()
            self.assertPathDoesNotExist("blah")

    def test_tree_path(self):
        tree = self.make_branch_and_tree(".", format=self._format)
        tree.mkdir("subdir")
        tree.commit("Add subdir")
        with Workspace(tree, use_inotify=self._use_inotify) as ws:
            self.assertEqual("foo", ws.tree_path("foo"))
            self.assertEqual("", ws.tree_path())
        with Workspace(tree, subpath="subdir", use_inotify=self._use_inotify) as ws:
            self.assertEqual("subdir/foo", ws.tree_path("foo"))
            self.assertEqual("subdir/", ws.tree_path())

    def test_abspath(self):
        tree = self.make_branch_and_tree(".", format=self._format)
        tree.mkdir("subdir")
        tree.commit("Add subdir")
        with Workspace(tree, use_inotify=self._use_inotify) as ws:
            self.assertEqual(tree.abspath("foo"), ws.abspath("foo"))
            self.assertEqual(tree.abspath(""), ws.abspath())
        with Workspace(tree, subpath="subdir", use_inotify=self._use_inotify) as ws:
            self.assertEqual(tree.abspath("subdir/foo"), ws.abspath("foo"))
            self.assertEqual(tree.abspath("subdir") + "/", ws.abspath(""))
            self.assertEqual(tree.abspath("subdir") + "/", ws.abspath())

    def test_open_containing(self):
        tree = self.make_branch_and_tree(".", format=self._format)
        tree.mkdir("subdir")
        tree.commit("Add subdir")
        ws = Workspace.from_path("subdir")
        self.assertEqual(ws.tree.abspath("."), tree.abspath("."))
        self.assertEqual(ws.subpath, "subdir")
        self.assertEqual(None, ws.use_inotify)



#!/usr/bin/python3

"""Cross-platform os tools: files/directories manipulations
Usage:

    ostools.py help
                    prints this help

    ostools.py copytodir FILES... DIR
                    copy files to specified directory

    ostools.py copytree FILES... DIR
                    copy files to specified directory keeping relative paths

    ostools.py copydir SOURCE TARGET
                    recursively copy SOURCE directory tree to TARGET directory

    ostools.py remove [FILES...] [DIRS...]
                    remove files or directories (recursive)
"""

import glob
import os
import shutil
import sys


def makedir(dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    if not os.path.isdir(dirname):
        print("Error: Destination is not a directory", dirname)
        return 2
    return 0


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    if not argv:
        argv = ["help"]

    cmd = argv.pop(0)

    if cmd == "help":
        print(__doc__)
        return 0

    if cmd == "copytodir":
        if len(argv) < 2:
            print("Usage:  ostools.py copytodir FILES... DIR")
            return 1

        todir = argv.pop()
        retcode = makedir(todir)
        if retcode:
            return retcode

        files = []
        for possible_glob in argv:
            files += glob.glob(possible_glob)

        for src in files:
            dest = os.path.join(todir, os.path.basename(src))
            shutil.copy(src, dest)
            print("Copied:", src, "=>", dest)

        return 0

    if cmd == "copytree":
        if len(argv) < 2:
            print("Usage:  ostools.py copytree FILES... DIR")
            return 1

        todir = argv.pop()
        retcode = makedir(todir)
        if retcode:
            return retcode

        files = []
        for possible_glob in argv:
            files += glob.glob(possible_glob)

        for src in files:
            relative_path = src
            dest = os.path.join(todir, relative_path)
            dest_dir = os.path.dirname(dest)
            retcode = makedir(dest_dir)
            if retcode:
                return retcode
            shutil.copy(src, dest)
            print("Copied:", src, "=>", dest)

        return 0

    if cmd == "copydir":
        if len(argv) != 2:
            print("Usage:  ostools.py copydir SOURCE TARGET")
            return 1

        def _copy(src, dest, follow_symlinks=True):
            shutil.copy(src, dest, follow_symlinks=follow_symlinks)
            print("Copied:", src, "=>", dest)

        shutil.copytree(argv[0], argv[1], copy_function=_copy, dirs_exist_ok=True)

        return 0

    if cmd == "remove":
        if len(argv) == 0:
            print("Usage:  ostools.py remove [FILES...] [DIRS...]")
            return 1

        filesdirs = []
        for possible_glob in argv:
            filesdirs += glob.glob(possible_glob)

        for i in filesdirs:
            if os.path.isdir(i):
                shutil.rmtree(i)
                print("Removed:", i)
            elif os.path.isfile(i):
                os.remove(i)
                print("Removed:", i)
            else:
                print("Not found:", i)

        return 0

    if cmd == "basename":
        if len(argv) == 0:
            print("Usage:  ostools.py basename [PATH | URL]")
            return 1

        for path in argv:
            print(os.path.basename(path))
        return 0

    if cmd == "makedir":
        if len(argv) == 0:
            print("Usage:  ostools.py makedir DIR")
            return 1

        retcode = makedir(argv.pop())
        if retcode:
            return retcode
        return 0

    print("Usage error")
    print(__doc__)
    return 1


if __name__ == "__main__":
    sys.exit(main())



# Copyright (C) 2006, 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Tests of the 'brz add' command."""

import os

from breezy import osutils, tests
from breezy.tests import features, script
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class TestAdd(tests.TestCaseWithTransport):
    scenarios = [
        ("pre-views", {"branch_tree_format": "pack-0.92"}),
        ("view-aware", {"branch_tree_format": "2a"}),
    ]

    def make_branch_and_tree(self, dir):
        return super().make_branch_and_tree(dir, format=self.branch_tree_format)

    def test_add_reports(self):
        """Add command prints the names of added files."""
        self.make_branch_and_tree(".")
        self.build_tree(["top.txt", "dir/", "dir/sub.txt", "CVS"])
        self.build_tree_contents([(".bzrignore", b"CVS\n")])
        out = self.run_bzr("add")[0]
        # the ordering is not defined at the moment
        results = sorted(out.rstrip("\n").split("\n"))
        self.assertEqual(
            ["adding .bzrignore", "adding dir", "adding dir/sub.txt", "adding top.txt"],
            results,
        )
        out = self.run_bzr("add -v")[0]
        results = sorted(out.rstrip("\n").split("\n"))
        self.assertEqual(['ignored CVS matching "CVS"'], results)

    def test_add_quiet_is(self):
        """Add -q does not print the names of added files."""
        self.make_branch_and_tree(".")
        self.build_tree(["top.txt", "dir/", "dir/sub.txt"])
        out = self.run_bzr("add -q")[0]
        # the ordering is not defined at the moment
        results = sorted(out.rstrip("\n").split("\n"))
        self.assertEqual([""], results)

    def test_add_in_unversioned(self):
        """Try to add a file in an unversioned directory.

        "brz add" should add the parent(s) as necessary.
        """
        self.make_branch_and_tree(".")
        self.build_tree(["inertiatic/", "inertiatic/esp"])
        self.assertEqual(self.run_bzr("unknowns")[0], "inertiatic\n")
        self.run_bzr("add inertiatic/esp")
        self.assertEqual(self.run_bzr("unknowns")[0], "")

        # Multiple unversioned parents
        self.build_tree(["veil/", "veil/cerpin/", "veil/cerpin/taxt"])
        self.assertEqual(self.run_bzr("unknowns")[0], "veil\n")
        self.run_bzr("add veil/cerpin/taxt")
        self.assertEqual(self.run_bzr("unknowns")[0], "")

        # Check whacky paths work
        self.build_tree(["cicatriz/", "cicatriz/esp"])
        self.assertEqual(self.run_bzr("unknowns")[0], "cicatriz\n")
        self.run_bzr("add inertiatic/../cicatriz/esp")
        self.assertEqual(self.run_bzr("unknowns")[0], "")

    def test_add_no_recurse(self):
        self.make_branch_and_tree(".")
        self.build_tree(["inertiatic/", "inertiatic/esp"])
        self.assertEqual(self.run_bzr("unknowns")[0], "inertiatic\n")
        self.run_bzr("add -N inertiatic")
        self.assertEqual(self.run_bzr("unknowns")[0], "inertiatic/esp\n")

    def test_add_in_versioned(self):
        """Try to add a file in a versioned directory.

        "brz add" should do this happily.
        """
        self.make_branch_and_tree(".")
        self.build_tree(["inertiatic/", "inertiatic/esp"])
        self.assertEqual(self.run_bzr("unknowns")[0], "inertiatic\n")
        self.run_bzr("add --no-recurse inertiatic")
        self.assertEqual(self.run_bzr("unknowns")[0], "inertiatic/esp\n")
        self.run_bzr("add inertiatic/esp")
        self.assertEqual(self.run_bzr("unknowns")[0], "")

    def test_subdir_add(self):
        """Add in subdirectory should add only things from there down."""
        eq = self.assertEqual

        t = self.make_branch_and_tree(".")
        self.build_tree(["src/", "README"])

        eq(sorted(t.unknowns()), ["README", "src"])

        self.run_bzr("add src")

        self.build_tree(["src/foo.c"])

        # add with no arguments in a subdirectory gets only files below that
        # subdirectory
        self.run_bzr("add", working_dir="src")
        self.assertEqual("README\n", self.run_bzr("unknowns", working_dir="src")[0])
        # reopen to see the new changes
        t = t.controldir.open_workingtree("src")
        versioned = [path for path, entry in t.iter_entries_by_dir()]
        self.assertEqual(versioned, ["", "src", "src/foo.c"])

        # add from the parent directory should pick up all file names
        self.run_bzr("add")
        self.assertEqual(self.run_bzr("unknowns")[0], "")
        self.run_bzr("check")

    def test_add_missing(self):
        """Brz add foo where foo is missing should error."""
        self.make_branch_and_tree(".")
        self.run_bzr("add missing-file", retcode=3)

    def test_add_from(self):
        base_tree = self.make_branch_and_tree("base")
        self.build_tree(["base/a", "base/b/", "base/b/c"])
        base_tree.add(["a", "b", "b/c"])
        base_tree.commit("foo")

        new_tree = self.make_branch_and_tree("new")
        self.build_tree(["new/a", "new/b/", "new/b/c", "d"])

        out, err = self.run_bzr("add --file-ids-from ../base", working_dir="new")
        self.assertEqual("", err)
        self.assertEqualDiff(
            "adding a w/ file id from a\n"
            "adding b w/ file id from b\n"
            "adding b/c w/ file id from b/c\n",
            out,
        )
        new_tree = new_tree.controldir.open_workingtree()
        self.assertEqual(base_tree.path2id("a"), new_tree.path2id("a"))
        self.assertEqual(base_tree.path2id("b"), new_tree.path2id("b"))
        self.assertEqual(base_tree.path2id("b/c"), new_tree.path2id("b/c"))

    def test_add_from_subdir(self):
        base_tree = self.make_branch_and_tree("base")
        self.build_tree(["base/a", "base/b/", "base/b/c", "base/b/d"])
        base_tree.add(["a", "b", "b/c", "b/d"])
        base_tree.commit("foo")

        new_tree = self.make_branch_and_tree("new")
        self.build_tree(["new/c", "new/d"])

        out, err = self.run_bzr("add --file-ids-from ../base/b", working_dir="new")
        self.assertEqual("", err)
        self.assertEqualDiff(
            "adding c w/ file id from b/c\nadding d w/ file id from b/d\n", out
        )

        new_tree = new_tree.controldir.open_workingtree("new")
        self.assertEqual(base_tree.path2id("b/c"), new_tree.path2id("c"))
        self.assertEqual(base_tree.path2id("b/d"), new_tree.path2id("d"))

    def test_add_dry_run(self):
        """Test a dry run add, make sure nothing is added."""
        wt = self.make_branch_and_tree(".")
        self.build_tree(["inertiatic/", "inertiatic/esp"])
        self.assertEqual(list(wt.unknowns()), ["inertiatic"])
        self.run_bzr("add --dry-run")
        self.assertEqual(list(wt.unknowns()), ["inertiatic"])

    def test_add_control_dir(self):
        """The control dir and its content should be refused."""
        self.make_branch_and_tree(".")
        err = self.run_bzr("add .bzr", retcode=3)[1]
        self.assertContainsRe(err, r"ERROR:.*\.bzr.*control file")
        err = self.run_bzr("add .bzr/README", retcode=3)[1]
        self.assertContainsRe(err, r"ERROR:.*\.bzr.*control file")
        self.build_tree([".bzr/crescent"])
        err = self.run_bzr("add .bzr/crescent", retcode=3)[1]
        self.assertContainsRe(err, r"ERROR:.*\.bzr.*control file")

    def test_add_via_symlink(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        self.make_branch_and_tree("source")
        self.build_tree(["source/top.txt"])
        os.symlink("source", "link")
        out = self.run_bzr(["add", "link/top.txt"])[0]
        self.assertEqual(out, "adding top.txt\n")

    def test_add_symlink_to_abspath(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        self.make_branch_and_tree("tree")
        os.symlink(osutils.abspath("target"), "tree/link")
        out = self.run_bzr(["add", "tree/link"])[0]
        self.assertEqual(out, "adding link\n")

    def test_add_not_child(self):
        # https://bugs.launchpad.net/bzr/+bug/98735
        sr = script.ScriptRunner()
        self.make_branch_and_tree("tree1")
        self.make_branch_and_tree("tree2")
        self.build_tree(["tree1/a", "tree2/b"])
        sr.run_script(
            self,
            """
        $ brz add tree1/a tree2/b
        2>brz: ERROR: Path "...tree2/b" is not a child of path "...tree1"
        """,
        )

    def test_add_multiple_files_in_unicode_cwd(self):
        """Adding multiple files in a non-ascii cwd, see lp:686611."""
        self.requireFeature(features.UnicodeFilenameFeature)
        self.make_branch_and_tree("\xa7")
        self.build_tree(["\xa7/a", "\xa7/b"])
        out, err = self.run_bzr(["add", "a", "b"], working_dir="\xa7")
        self.assertEqual(out, "adding a\nadding b\n")
        self.assertEqual(err, "")

    def test_add_skip_large_files(self):
        """Test skipping files larger than add.maximum_file_size."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["small.txt", "big.txt", "big2.txt"])
        self.build_tree_contents([("small.txt", b"0\n")])
        self.build_tree_contents([("big.txt", b"01234567890123456789\n")])
        self.build_tree_contents([("big2.txt", b"01234567890123456789\n")])
        tree.branch.get_config_stack().set("add.maximum_file_size", 5)
        out = self.run_bzr("add")[0]
        results = sorted(out.rstrip("\n").split("\n"))
        self.assertEqual(["adding small.txt"], results)
        # named items never skipped, even if over max
        out, err = self.run_bzr(["add", "big2.txt"])
        results = sorted(out.rstrip("\n").split("\n"))
        self.assertEqual(["adding big2.txt"], results)
        self.assertEqual("", err)
        tree.branch.get_config_stack().set("add.maximum_file_size", 30)
        out = self.run_bzr("add")[0]
        results = sorted(out.rstrip("\n").split("\n"))
        self.assertEqual(["adding big.txt"], results)

    def test_add_backslash(self):
        # pad.lv/165151
        if os.path.sep == "\\":
            # TODO(jelmer): Test that backslashes are appropriately
            # ignored?
            raise tests.TestNotApplicable(
                "unable to add filenames with backslashes where "
                " it is the path separator"
            )
        self.make_branch_and_tree(".")
        self.build_tree(["\\"])
        self.assertEqual("adding \\\n", self.run_bzr("add \\\\")[0])
        self.assertEqual("\\\n", self.run_bzr("ls --versioned")[0])



# Copyright (C) 2019 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy.tests import TestCaseWithTransport


class TestCatRevision(TestCaseWithTransport):
    def test_already_url(self):
        wt = self.make_branch_and_tree(".")

        out, err = self.run_bzr("resolve-location {}".format(wt.branch.user_url))
        self.assertEqual(out, "{}\n".format(wt.branch.user_url.replace("file://", "")))

    def test_parent_missing(self):
        self.make_branch_and_tree(".")

        out, err = self.run_bzr("resolve-location :parent", retcode=3)
        self.assertEqual(out, "")
        self.assertEqual(err, "brz: ERROR: No parent location assigned.\n")

    def test_parent(self):
        wt = self.make_branch_and_tree(".")
        wt.branch.set_parent("http://example.com/foo")

        out, err = self.run_bzr("resolve-location :parent")
        self.assertEqual(out, "http://example.com/foo\n")
        self.assertEqual(err, "")



# Copyright (C) 2005, 2006, 2007, 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import sys

import breezy.errors
from breezy import urlutils
from breezy.osutils import getcwd
from breezy.tests import TestCaseWithTransport, TestNotApplicable, TestSkipped

"""Tests for Branch parent URL"""


class TestParent(TestCaseWithTransport):
    def test_no_default_parent(self):
        """Branches should have no parent by default."""
        b = self.make_branch(".")
        self.assertEqual(None, b.get_parent())

    def test_set_get_parent(self):
        """Set, re-get and reset the parent."""
        b = self.make_branch("subdir")
        url = "http://example.com/bzr/bzr.dev"
        b.set_parent(url)
        self.assertEqual(url, b.get_parent())
        self.assertEqual(url, b._get_parent_location())

        b.set_parent(None)
        self.assertEqual(None, b.get_parent())

        b.set_parent("../other_branch")

        expected_parent = urlutils.join(self.get_url("subdir"), "../other_branch")
        self.assertEqual(expected_parent, b.get_parent())
        path = urlutils.join(self.get_url("subdir"), "../yanb")
        b.set_parent(path)
        self.assertEqual("../yanb", b._get_parent_location())
        self.assertEqual(path, b.get_parent())

        self.assertRaises(urlutils.InvalidURL, b.set_parent, "\xb5")
        b.set_parent(urlutils.escape("\xb5"))
        self.assertEqual("%C2%B5", b._get_parent_location())

        self.assertEqual(b.base + "%C2%B5", b.get_parent())

        # Handle the case for older style absolute local paths
        if sys.platform == "win32":
            # TODO: jam 20060515 Do we want to special case Windows local
            #       paths as well? Nobody has complained about it.
            pass
        else:
            b.lock_write()
            b._set_parent_location("/local/abs/path")
            b.unlock()
            self.assertEqual("file:///local/abs/path", b.get_parent())

    def test_get_invalid_parent(self):
        b = self.make_branch(".")

        cwd = getcwd()
        n_dirs = len(cwd.split("/"))

        # Force the relative path to be something invalid
        # This should attempt to go outside the filesystem
        path = ("../" * (n_dirs + 5)) + "foo"
        b.lock_write()
        b._set_parent_location(path)
        b.unlock()

        # With an invalid branch parent, just return None
        self.assertRaises(breezy.errors.InaccessibleParent, b.get_parent)

    def test_win32_set_parent_on_another_drive(self):
        if sys.platform != "win32":
            raise TestSkipped("windows-specific test")
        b = self.make_branch(".")
        base_url = b.controldir.transport.abspath(".")
        if not base_url.startswith("file:///"):
            raise TestNotApplicable("this test should be run with local base")
        base = urlutils.local_path_from_url(base_url)
        other = "file:///D:/path"
        if base[0] != "C":
            other = "file:///C:/path"
        b.set_parent(other)
        self.assertEqual(other, b._get_parent_location())



#
# Breezy documentation build configuration file, created by
# sphinx-quickstart on Tue Jul 21 17:04:52 2009.
#
# This file is execfile()d with the current directory set to its containing dir.

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path = [os.path.abspath("../..")] + sys.path

# Most of the configuration for Breezy docs is defined here ...
from breezy.doc_generate.conf import *

## Configuration specific to this site ##

# The locale code for this documentation set
brz_locale = "en"

# Translations & supporting helper function
brz_titles = {
    "Table of Contents (%s)": None,
    "Breezy User Guide": None,
    "Breezy User Reference": None,
    "Breezy Release Notes": None,
    "Breezy Upgrade Guide": None,
    "Breezy System Administrator's Guide": None,
    "Breezy in five minutes": None,
    "Breezy Tutorial": None,
    "Using Breezy With Launchpad": None,
    "Centralized Workflow Tutorial": None,
    "What's New in Breezy 2.1?": None,
}


def brz_title(s):
    return brz_titles.get(s) or s


# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = brz_title("Table of Contents (%s)") % (release,)

# Additional templates that should be rendered to pages, maps page names to
# template names.
html_additional_pages = {"index": "index.html"}

# Output file base name for HTML help builder.
htmlhelp_basename = "brz-{}".format(brz_locale)

# Grouping the document tree into files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
brz_documents = [
    # Manuals
    (
        "user-guide/index",
        "brz-{}-user-guide".format(brz_locale),
        brz_title("Breezy User Guide"),
        brz_team,
        "manual",
    ),
    (
        "user-reference/index",
        "brz-{}-user-reference".format(brz_locale),
        brz_title("Breezy User Reference"),
        brz_team,
        "manual",
    ),
    (
        "release-notes/index",
        "brz-{}-release-notes".format(brz_locale),
        brz_title("Breezy Release Notes"),
        brz_team,
        "manual",
    ),
    (
        "upgrade-guide/index",
        "brz-{}-upgrade-guide".format(brz_locale),
        brz_title("Breezy Upgrade Guide"),
        brz_team,
        "manual",
    ),
    (
        "admin-guide/index",
        "brz-{}-admin-guide".format(brz_locale),
        brz_title("Breezy System Administrator's Guide"),
        brz_team,
        "manual",
    ),
    # Tutorials
    (
        "mini-tutorial/index",
        "brz-{}-tutorial-mini".format(brz_locale),
        brz_title("Breezy in five minutes"),
        brz_team,
        "howto",
    ),
    (
        "tutorials/tutorial",
        "brz-{}-tutorial".format(brz_locale),
        brz_title("Breezy Tutorial"),
        brz_team,
        "howto",
    ),
    (
        "tutorials/using_breezy_with_github",
        "brz-{}-tutorial-with-github".format(brz_locale),
        brz_title("Using Breezy With GitHub"),
        brz_team,
        "howto",
    ),
    (
        "tutorials/using_breezy_with_launchpad",
        "brz-{}-tutorial-with-launchpad".format(brz_locale),
        brz_title("Using Breezy With Launchpad"),
        brz_team,
        "howto",
    ),
    (
        "tutorials/centralized_workflow",
        "brz-{}-tutorial-centralized".format(brz_locale),
        brz_title("Centralized Workflow Tutorial"),
        brz_team,
        "howto",
    ),
    (
        "whats-new/whats-new-in-3.1",
        "brz-{}-whats-new".format(brz_locale),
        brz_title("What's New in Breezy 3.1?"),
        brz_team,
        "howto",
    ),
]

latex_documents = [
    (start, target + ".tex", title, author, doc_class)
    for start, target, title, author, doc_class in brz_documents
]

texinfo_documents = [
    (start, target, title, author, doc_class)
    for start, target, title, author, doc_class in brz_documents
]

# List of documents that shouldn't be included in the build.
unused_docs = [
    # Subtopics that get included
    "upgrade-guide/overview",
    "upgrade-guide/data_migration",
    "upgrade-guide/tips_and_tricks",
    # Plain-style documentation generation stuff
    "release-notes/NEWS",
    "user-reference/brz_man",
    "user-guide/index-plain",
    "admin-guide/index-plain",
    # Templates
    "release-notes/release-template",
    "release-notes/series-template",
    # Miscellaneous
    "user-reference/readme",
]



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

from .... import tests
from ..helpers import kind_to_mode
from . import FastimportFeature

try:
    from fastimport import commands
except ImportError:
    commands = object()  # type: ignore


def load_tests(loader, standard_tests, pattern):
    """Parameterize tests for all versions of groupcompress."""
    scenarios = [
        ("pack-0.92", {"branch_format": "pack-0.92"}),
        ("1.9-rich-root", {"branch_format": "1.9-rich-root"}),
    ]

    scenarios.append(("2a", {"branch_format": "2a"}))
    suite = loader.suiteClass()
    result = tests.multiply_tests(standard_tests, scenarios, suite)
    return result


class TestCaseForGenericProcessor(tests.TestCaseWithTransport):
    _test_needs_features = [FastimportFeature]

    branch_format = "pack-0.92"

    def get_handler(self):
        from ..processors import generic_processor

        branch = self.make_branch(".", format=self.branch_format)
        handler = generic_processor.GenericProcessor(branch.controldir)
        return handler, branch

    # FIXME: [] as a default is bad, as it is mutable, but I want
    # to use None to mean "don't check this".
    def assertChanges(
        self,
        branch,
        revno,
        expected_added=None,
        expected_removed=None,
        expected_modified=None,
        expected_renamed=None,
        expected_kind_changed=None,
    ):
        """Check the changes introduced in a revision of a branch.

        This method checks that a revision introduces expected changes.
        The required changes are passed in as a list, where
        each entry contains the needed information about the change.

        If you do not wish to assert anything about a particular
        category then pass None instead.

        branch: The branch.
        revno: revision number of revision to check.
        expected_added: a list of (filename,) tuples that must have
            been added in the delta.
        expected_removed: a list of (filename,) tuples that must have
            been removed in the delta.
        expected_modified: a list of (filename,) tuples that must have
            been modified in the delta.
        expected_renamed: a list of (old_path, new_path) tuples that
            must have been renamed in the delta.
        expected_kind_changed: a list of (path, old_kind, new_kind) tuples
            that must have been changed in the delta.
        :return: revtree1, revtree2
        """
        if expected_kind_changed is None:
            expected_kind_changed = []
        if expected_renamed is None:
            expected_renamed = []
        if expected_modified is None:
            expected_modified = []
        if expected_removed is None:
            expected_removed = []
        if expected_added is None:
            expected_added = []
        repo = branch.repository
        revtree1 = repo.revision_tree(branch.get_rev_id(revno - 1))
        revtree2 = repo.revision_tree(branch.get_rev_id(revno))
        changes = revtree2.changes_from(revtree1)
        self._check_changes(
            changes,
            expected_added,
            expected_removed,
            expected_modified,
            expected_renamed,
            expected_kind_changed,
        )
        return revtree1, revtree2

    def _check_changes(
        self,
        changes,
        expected_added=None,
        expected_removed=None,
        expected_modified=None,
        expected_renamed=None,
        expected_kind_changed=None,
    ):
        """Check the changes in a TreeDelta.

        This method checks that the TreeDelta contains the expected
        modifications between the two trees that were used to generate
        it. The required changes are passed in as a list, where
        each entry contains the needed information about the change.

        If you do not wish to assert anything about a particular
        category then pass None instead.

        changes: The TreeDelta to check.
        expected_added: a list of (filename,) tuples that must have
            been added in the delta.
        expected_removed: a list of (filename,) tuples that must have
            been removed in the delta.
        expected_modified: a list of (filename,) tuples that must have
            been modified in the delta.
        expected_renamed: a list of (old_path, new_path) tuples that
            must have been renamed in the delta.
        expected_kind_changed: a list of (path, old_kind, new_kind) tuples
            that must have been changed in the delta.
        """
        if expected_kind_changed is None:
            expected_kind_changed = []
        if expected_renamed is None:
            expected_renamed = []
        if expected_modified is None:
            expected_modified = []
        if expected_removed is None:
            expected_removed = []
        if expected_added is None:
            expected_added = []
        renamed = changes.renamed
        added = changes.added + changes.copied
        removed = changes.removed
        modified = changes.modified
        kind_changed = changes.kind_changed
        if expected_renamed is not None:
            self.assertEqual(
                len(renamed),
                len(expected_renamed),
                "{} is renamed, expected {}".format(renamed, expected_renamed),
            )
            renamed_files = [(item.path[0], item.path[1]) for item in renamed]
            for expected_renamed_entry in expected_renamed:
                expected_renamed_entry = (
                    expected_renamed_entry[0].decode("utf-8"),
                    expected_renamed_entry[1].decode("utf-8"),
                )
                self.assertTrue(
                    expected_renamed_entry in renamed_files,
                    "{} is not renamed, {} are".format(
                        expected_renamed_entry, renamed_files
                    ),
                )
        if expected_added is not None:
            self.assertEqual(
                len(added), len(expected_added), "{} is added".format(str(added))
            )
            added_files = [(item.path[1],) for item in added]
            for expected_added_entry in expected_added:
                expected_added_entry = (expected_added_entry[0].decode("utf-8"),)
                self.assertTrue(
                    expected_added_entry in added_files,
                    "{} is not added, {} are".format(expected_added_entry, added_files),
                )
        if expected_removed is not None:
            self.assertEqual(
                len(removed),
                len(expected_removed),
                "{} is removed".format(str(removed)),
            )
            removed_files = [(item.path[0],) for item in removed]
            for expected_removed_entry in expected_removed:
                expected_removed_entry = (expected_removed_entry[0].decode("utf-8"),)
                self.assertTrue(
                    expected_removed_entry in removed_files,
                    "{} is not removed, {} are".format(
                        expected_removed_entry, removed_files
                    ),
                )
        if expected_modified is not None:
            self.assertEqual(
                len(modified),
                len(expected_modified),
                "{} is modified".format(str(modified)),
            )
            modified_files = [(item.path[1],) for item in modified]
            for expected_modified_entry in expected_modified:
                expected_modified_entry = (expected_modified_entry[0].decode("utf-8"),)
                self.assertTrue(
                    expected_modified_entry in modified_files,
                    "{} is not modified, {} are".format(
                        expected_modified_entry, modified_files
                    ),
                )
        if expected_kind_changed is not None:
            self.assertEqual(
                len(kind_changed),
                len(expected_kind_changed),
                "{} is kind-changed, expected {}".format(
                    kind_changed, expected_kind_changed
                ),
            )
            kind_changed_files = [
                (item.path[1], item.kind[0], item.kind[1]) for item in kind_changed
            ]
            for expected_kind_changed_entry in expected_kind_changed:
                expected_kind_changed_entry = (
                    expected_kind_changed_entry[0].decode("utf-8"),
                ) + expected_kind_changed_entry[1:]
                self.assertTrue(
                    expected_kind_changed_entry in kind_changed_files,
                    "{} is not kind-changed, {} are".format(
                        expected_kind_changed_entry, kind_changed_files
                    ),
                )

    def assertContent(self, branch, tree, path, content):
        with branch.lock_read():
            self.assertEqual(tree.get_file_text(path.decode("utf-8")), content)

    def assertSymlinkTarget(self, branch, tree, path, target):
        with branch.lock_read():
            self.assertEqual(tree.get_symlink_target(path.decode("utf-8")), target)

    def assertExecutable(self, branch, tree, path, executable):
        with branch.lock_read():
            self.assertEqual(tree.is_executable(path.decode("utf-8")), executable)

    def assertRevisionRoot(self, revtree, path):
        self.assertEqual(
            revtree.get_revision_id(), revtree.get_file_revision(path.decode("utf-8"))
        )


class TestImportToPackTag(TestCaseForGenericProcessor):
    def file_command_iter(
        self,
        path,
        kind="file",
        content=b"aaa",
        executable=False,
        to_kind=None,
        to_content=b"bbb",
        to_executable=None,
    ):
        # Revno 1: create a file or symlink
        # Revno 2: modify it
        if to_kind is None:
            to_kind = kind
        if to_executable is None:
            to_executable = executable

        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    path, kind_to_mode(kind, executable), None, content
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileModifyCommand(
                    path, kind_to_mode(to_kind, to_executable), None, to_content
                )

            # pass "head" for from_ to show that #401249 is worse than I knew
            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b"head", [], files_two
            )

            yield commands.TagCommand(b"tag1", b":1", committer, b"tag 1")

            # pass "head" for from_ to demonstrate #401249
            yield commands.TagCommand(b"tag2", b"head", committer, b"tag 2")

        return command_list

    def test_tag(self):
        handler, branch = self.get_handler()
        path = b"a"
        raise tests.KnownFailure("non-mark committish not yet supported- bug #410249")
        handler.process(self.file_command_iter(path))


class TestImportZeroMarker(TestCaseForGenericProcessor):
    def test_tag(self):
        handler, branch = self.get_handler()

        def command_list():
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]
            yield commands.TagCommand(b"tag1", b":0", committer, b"tag 1")

        handler.process(command_list)


class TestImportToPackModify(TestCaseForGenericProcessor):
    def file_command_iter(
        self,
        path,
        kind="file",
        content=b"aaa",
        executable=False,
        to_kind=None,
        to_content=b"bbb",
        to_executable=None,
    ):
        # Revno 1: create a file or symlink
        # Revno 2: modify it
        if to_kind is None:
            to_kind = kind
        if to_executable is None:
            to_executable = executable
        mode = kind_to_mode(kind, executable)
        to_mode = kind_to_mode(to_kind, to_executable)

        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(path, mode, None, content)

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileModifyCommand(path, to_mode, None, to_content)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_modify_file_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2, expected_modified=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertContent(branch, revtree2, path, b"bbb")
        self.assertRevisionRoot(revtree1, path)
        self.assertRevisionRoot(revtree2, path)

    def test_modify_file_in_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(branch, 2, expected_modified=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertContent(branch, revtree2, path, b"bbb")

    def test_modify_symlink_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path, kind="symlink"))
        revtree1, revtree2 = self.assertChanges(branch, 2, expected_modified=[(path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")
        self.assertRevisionRoot(revtree1, path)
        self.assertRevisionRoot(revtree2, path)

    def test_modify_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path, kind="symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(branch, 2, expected_modified=[(path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")

    def test_modify_file_becomes_symlink(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path, kind="file", to_kind="symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_kind_changed=[(path, "file", "symlink")]
        )
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")

    def test_modify_symlink_becomes_file(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path, kind="symlink", to_kind="file"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_kind_changed=[(path, "symlink", "file")]
        )
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertContent(branch, revtree2, path, b"bbb")

    def test_modify_file_now_executable(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(
            self.file_command_iter(
                path, executable=False, to_executable=True, to_content=b"aaa"
            )
        )
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(branch, 2, expected_modified=[(path,)])
        self.assertExecutable(branch, revtree1, path, False)
        self.assertExecutable(branch, revtree2, path, True)

    def test_modify_file_no_longer_executable(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(
            self.file_command_iter(
                path, executable=True, to_executable=False, to_content=b"aaa"
            )
        )
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(branch, 2, expected_modified=[(path,)])
        self.assertExecutable(branch, revtree1, path, True)
        self.assertExecutable(branch, revtree2, path, False)


class TestImportToPackModifyTwice(TestCaseForGenericProcessor):
    """This tests when the same file is modified twice in the one commit.

    Note: hg-fast-export produces data like this on occasions.
    """

    def file_command_iter(
        self,
        path,
        kind="file",
        content=b"aaa",
        executable=False,
        to_kind=None,
        to_content=b"bbb",
        to_executable=None,
    ):
        # Revno 1: create a file twice
        if to_kind is None:
            to_kind = kind
        if to_executable is None:
            to_executable = executable

        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    path, kind_to_mode(kind, executable), None, content
                )
                yield commands.FileModifyCommand(
                    path, kind_to_mode(to_kind, to_executable), None, to_content
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

        return command_list

    def test_modify_file_twice_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertRevisionRoot(revtree1, path)


class TestImportToPackModifyTricky(TestCaseForGenericProcessor):
    def file_command_iter(self, path1, path2, kind="file"):
        # Revno 1: create a file or symlink in a directory
        # Revno 2: create a second file that implicitly deletes the
        # first one because either:
        # * the new file is a in directory with the old file name
        # * the new file has the same name as the directory of the first
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    path1, kind_to_mode(kind, False), None, b"aaa"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileModifyCommand(
                    path2, kind_to_mode(kind, False), None, b"bbb"
                )

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_modify_file_becomes_directory(self):
        handler, branch = self.get_handler()
        path1 = b"a/b"
        path2 = b"a/b/c"
        handler.process(self.file_command_iter(path1, path2))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path1,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_added=[(path2,)],
            expected_kind_changed=[(path1, "file", "directory")],
        )
        self.assertContent(branch, revtree1, path1, b"aaa")
        self.assertContent(branch, revtree2, path2, b"bbb")

    def test_modify_directory_becomes_file(self):
        handler, branch = self.get_handler()
        path1 = b"a/b/c"
        path2 = b"a/b"
        handler.process(self.file_command_iter(path1, path2))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (b"a/b",), (path1,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[
                (path1,),
            ],
            expected_kind_changed=[(path2, "directory", "file")],
        )
        self.assertContent(branch, revtree1, path1, b"aaa")
        self.assertContent(branch, revtree2, path2, b"bbb")

    def test_modify_symlink_becomes_directory(self):
        handler, branch = self.get_handler()
        path1 = b"a/b"
        path2 = b"a/b/c"
        handler.process(self.file_command_iter(path1, path2, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path1,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_added=[(path2,)],
            expected_kind_changed=[(path1, "symlink", "directory")],
        )
        self.assertSymlinkTarget(branch, revtree1, path1, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path2, "bbb")

    def test_modify_directory_becomes_symlink(self):
        handler, branch = self.get_handler()
        path1 = b"a/b/c"
        path2 = b"a/b"
        handler.process(self.file_command_iter(path1, path2, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (b"a/b",), (path1,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[
                (path1,),
            ],
            expected_kind_changed=[(path2, "directory", "symlink")],
        )
        self.assertSymlinkTarget(branch, revtree1, path1, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path2, "bbb")


class TestImportToPackDelete(TestCaseForGenericProcessor):
    def file_command_iter(self, path, kind="file"):
        # Revno 1: create a file or symlink
        # Revno 2: delete it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    path, kind_to_mode(kind, False), None, b"aaa"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileDeleteCommand(path)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_delete_file_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(path,)])
        revtree1, revtree2 = self.assertChanges(branch, 2, expected_removed=[(path,)])
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertRevisionRoot(revtree1, path)

    def test_delete_file_in_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(b"a",), (path,)]
        )
        self.assertContent(branch, revtree1, path, b"aaa")

    def test_delete_symlink_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path, kind="symlink"))
        revtree1, revtree2 = self.assertChanges(branch, 2, expected_removed=[(path,)])
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertRevisionRoot(revtree1, path)

    def test_delete_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path, kind="symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(b"a",), (path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")

    def test_delete_file_in_deep_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/b/c/d"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (b"a/b",), (b"a/b/c",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(b"a",), (b"a/b",), (b"a/b/c",), (path,)]
        )
        self.assertContent(branch, revtree1, path, b"aaa")


class TestImportToPackDeleteNew(TestCaseForGenericProcessor):
    """Test deletion of a newly added file."""

    def file_command_iter(self, path, kind="file"):
        # Revno 1: create a file or symlink then delete it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    path, kind_to_mode(kind, False), None, b"aaa"
                )
                yield commands.FileDeleteCommand(path)

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

        return command_list

    def test_delete_new_file_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
        )

    def test_delete_new_file_in_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
        )

    def test_delete_new_symlink_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path, kind="symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
        )

    def test_delete_new_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path, kind="symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
        )

    def test_delete_new_file_in_deep_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/b/c/d"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
        )


class TestImportToPackDeleteMultiLevel(TestCaseForGenericProcessor):
    def file_command_iter(self, paths, paths_to_delete):
        # Revno 1: create multiple files
        # Revno 2: delete multiple files
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                for i, path in enumerate(paths):
                    yield commands.FileModifyCommand(
                        path, kind_to_mode("file", False), None, b"aaa%d" % i
                    )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                for path in paths_to_delete:
                    yield commands.FileDeleteCommand(path)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_delete_files_in_multiple_levels(self):
        handler, branch = self.get_handler()
        paths = [b"a/b/c", b"a/b/d/e"]
        paths_to_delete = [b"a/b/c", b"a/b/d/e"]
        handler.process(self.file_command_iter(paths, paths_to_delete))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
            expected_added=[
                (b"a",),
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
                (b"a/b/d/e",),
            ],
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[
                (b"a",),
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
                (b"a/b/d/e",),
            ],
        )

    def test_delete_file_single_level(self):
        handler, branch = self.get_handler()
        paths = [b"a/b/c", b"a/b/d/e"]
        paths_to_delete = [b"a/b/d/e"]
        handler.process(self.file_command_iter(paths, paths_to_delete))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
            expected_added=[
                (b"a",),
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
                (b"a/b/d/e",),
            ],
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[
                (b"a/b/d",),
                (b"a/b/d/e",),
            ],
        )

    def test_delete_file_complex_level(self):
        handler, branch = self.get_handler()
        paths = [b"a/b/c", b"a/b/d/e", b"a/f/g", b"a/h", b"a/b/d/i/j"]
        paths_to_delete = [b"a/b/c", b"a/b/d/e", b"a/f/g", b"a/b/d/i/j"]
        handler.process(self.file_command_iter(paths, paths_to_delete))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
            expected_added=[
                (b"a",),
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
                (b"a/b/d/e",),
                (b"a/f",),
                (b"a/f/g",),
                (b"a/h",),
                (b"a/b/d/i",),
                (b"a/b/d/i/j",),
            ],
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
                (b"a/b/d/e",),
                (b"a/f",),
                (b"a/f/g",),
                (b"a/b/d/i",),
                (b"a/b/d/i/j",),
            ],
        )


class TestImportToPackDeleteThenAdd(TestCaseForGenericProcessor):
    """Test delete followed by an add. Merges can cause this."""

    def file_command_iter(
        self,
        path,
        kind="file",
        content=b"aaa",
        executable=False,
        to_kind=None,
        to_content=b"bbb",
        to_executable=None,
    ):
        # Revno 1: create a file or symlink
        # Revno 2: delete it and add it
        if to_kind is None:
            to_kind = kind
        if to_executable is None:
            to_executable = executable

        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    path, kind_to_mode(kind, executable), None, content
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileDeleteCommand(path)
                yield commands.FileModifyCommand(
                    path, kind_to_mode(to_kind, to_executable), None, to_content
                )

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_delete_then_add_file_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(path,)])
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(path,)], expected_added=[(path,)]
        )
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertContent(branch, revtree2, path, b"bbb")
        self.assertRevisionRoot(revtree1, path)
        self.assertRevisionRoot(revtree2, path)

    def test_delete_then_add_file_in_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(path,)], expected_added=[(path,)]
        )
        self.assertContent(branch, revtree1, path, b"aaa")
        self.assertContent(branch, revtree2, path, b"bbb")

    def test_delete_then_add_symlink_in_root(self):
        handler, branch = self.get_handler()
        path = b"a"
        handler.process(self.file_command_iter(path, kind="symlink"))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(path,)], expected_added=[(path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")
        self.assertRevisionRoot(revtree1, path)
        self.assertRevisionRoot(revtree2, path)

    def test_delete_then_add_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        path = b"a/a"
        handler.process(self.file_command_iter(path, kind="symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(path,)], expected_added=[(path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, path, "bbb")


class TestImportToPackDeleteDirectory(TestCaseForGenericProcessor):
    def file_command_iter(self, paths, dir):
        # Revno 1: create multiple files
        # Revno 2: delete a directory holding those files
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                for i, path in enumerate(paths):
                    yield commands.FileModifyCommand(
                        path, kind_to_mode("file", False), None, b"aaa%d" % i
                    )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileDeleteCommand(dir)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_delete_dir(self):
        handler, branch = self.get_handler()
        paths = [b"a/b/c", b"a/b/d", b"a/b/e/f", b"a/g"]
        dir = b"a/b"
        handler.process(self.file_command_iter(paths, dir))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
            expected_added=[
                (b"a",),
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
                (b"a/b/e",),
                (b"a/b/e/f",),
                (b"a/g",),
            ],
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
                (b"a/b/e",),
                (b"a/b/e/f",),
            ],
        )


class TestImportToPackDeleteDirectoryThenAddFile(TestCaseForGenericProcessor):
    """Test deleting a directory then adding a file in the same commit."""

    def file_command_iter(self, paths, dir, new_path, kind="file"):
        # Revno 1: create files in a directory
        # Revno 2: delete the directory then add a file into it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                for i, path in enumerate(paths):
                    yield commands.FileModifyCommand(
                        path, kind_to_mode(kind, False), None, b"aaa%d" % i
                    )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileDeleteCommand(dir)
                yield commands.FileModifyCommand(
                    new_path, kind_to_mode(kind, False), None, b"bbb"
                )

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_delete_dir_then_add_file(self):
        handler, branch = self.get_handler()
        paths = [b"a/b/c", b"a/b/d"]
        dir = b"a/b"
        new_path = b"a/b/z"
        handler.process(self.file_command_iter(paths, dir, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
            expected_added=[
                (b"a",),
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
            ],
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(b"a/b",), (b"a/b/c",), (b"a/b/d",)],
            expected_added=[(b"a/b",), (b"a/b/z",)],
        )
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_delete_dir_then_add_symlink(self):
        handler, branch = self.get_handler()
        paths = [b"a/b/c", b"a/b/d"]
        dir = b"a/b"
        new_path = b"a/b/z"
        handler.process(self.file_command_iter(paths, dir, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch,
            1,
            expected_added=[
                (b"a",),
                (b"a/b",),
                (b"a/b/c",),
                (b"a/b/d",),
            ],
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(b"a/b",), (b"a/b/c",), (b"a/b/d",)],
            expected_added=[(b"a/b",), (b"a/b/z",)],
        )
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")


class TestImportToPackRename(TestCaseForGenericProcessor):
    def get_command_iter(self, old_path, new_path, kind="file"):
        # Revno 1: create a file or symlink
        # Revno 2: rename it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    old_path, kind_to_mode(kind, False), None, b"aaa"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileRenameCommand(old_path, new_path)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_rename_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"a/a"
        new_path = b"a/b"
        handler.process(self.get_command_iter(old_path, new_path))
        self.assertChanges(branch, 2, expected_renamed=[(old_path, new_path)])

    def test_rename_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"a/a"
        new_path = b"a/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        self.assertChanges(branch, 2, expected_renamed=[(old_path, new_path)])

    def test_rename_file_to_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"a/a"
        new_path = b"b/a"
        handler.process(self.get_command_iter(old_path, new_path))
        self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path, new_path)],
            expected_added=[(b"b",)],
            expected_removed=[(b"a",)],
        )

    def test_rename_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"a/a"
        new_path = b"b/a"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path, new_path)],
            expected_added=[(b"b",)],
            expected_removed=[(b"a",)],
        )


class TestImportToPackRenameNew(TestCaseForGenericProcessor):
    """Test rename of a newly added file."""

    def get_command_iter(self, old_path, new_path, kind="file"):
        # Revno 1: create a file and rename it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    old_path, kind_to_mode(kind, False), None, b"aaa"
                )
                yield commands.FileRenameCommand(old_path, new_path)

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

        return command_list

    def test_rename_new_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(new_path,)])
        self.assertRevisionRoot(revtree1, new_path)

    def test_rename_new_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(new_path,)])
        self.assertRevisionRoot(revtree1, new_path)

    def test_rename_new_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"a/a"
        new_path = b"a/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (new_path,)]
        )

    def test_rename_new_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"a/a"
        new_path = b"a/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (new_path,)]
        )


class TestImportToPackRenameToDeleted(TestCaseForGenericProcessor):
    """Test rename to a destination path deleted in this commit."""

    def get_command_iter(self, old_path, new_path, kind="file"):
        # Revno 1: create two files
        # Revno 2: delete one, rename the other one to that path
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    old_path, kind_to_mode(kind, False), None, b"aaa"
                )
                yield commands.FileModifyCommand(
                    new_path, kind_to_mode(kind, False), None, b"bbb"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileDeleteCommand(new_path)
                yield commands.FileRenameCommand(old_path, new_path)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_rename_to_deleted_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(old_path,), (new_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"bbb")
        self.assertContent(branch, revtree2, new_path, b"aaa")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree1, new_path)

    def test_rename_to_deleted_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(old_path,), (new_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, new_path, "aaa")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree1, new_path)

    def test_rename_to_deleted_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"d/a"
        new_path = b"d/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (old_path,), (new_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"bbb")
        self.assertContent(branch, revtree2, new_path, b"aaa")

    def test_rename_to_deleted_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"d/a"
        new_path = b"d/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (old_path,), (new_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, new_path, "aaa")

    def test_rename_to_deleted_file_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"d1/a"
        new_path = b"d2/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d1",), (old_path,), (b"d2",), (new_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(b"d1",), (new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"bbb")
        self.assertContent(branch, revtree2, new_path, b"aaa")

    def test_rename_to_deleted_symlink_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"d1/a"
        new_path = b"d2/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d1",), (old_path,), (b"d2",), (new_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(b"d1",), (new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, new_path, "aaa")


class TestImportToPackRenameModified(TestCaseForGenericProcessor):
    """Test rename of a path previously modified in this commit."""

    def get_command_iter(self, old_path, new_path, kind="file"):
        # Revno 1: create a file or symlink
        # Revno 2: modify then rename it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    old_path, kind_to_mode(kind, False), None, b"aaa"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileModifyCommand(
                    old_path, kind_to_mode(kind, False), None, b"bbb"
                )
                yield commands.FileRenameCommand(old_path, new_path)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_rename_of_modified_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_of_modified_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_of_modified_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"d/a"
        new_path = b"d/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (old_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_rename_of_modified_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"d/a"
        new_path = b"d/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (old_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")

    def test_rename_of_modified_file_to_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"d1/a"
        new_path = b"d2/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d1",), (old_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path, new_path)],
            expected_added=[(b"d2",)],
            expected_removed=[(b"d1",)],
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_rename_of_modified_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"d1/a"
        new_path = b"d2/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d1",), (old_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path, new_path)],
            expected_added=[(b"d2",)],
            expected_removed=[(b"d1",)],
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")


class TestImportToPackRenameThenModify(TestCaseForGenericProcessor):
    """Test rename of a path then modfy the new-path in the same commit."""

    def get_command_iter(self, old_path, new_path, kind="file"):
        # Revno 1: create a file or symlink
        # Revno 2: rename it then modify the newly created path
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    old_path, kind_to_mode(kind, False), None, b"aaa"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileRenameCommand(old_path, new_path)
                yield commands.FileModifyCommand(
                    new_path, kind_to_mode(kind, False), None, b"bbb"
                )

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_rename_then_modify_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_then_modify_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"d/a"
        new_path = b"d/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (old_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_rename_then_modify_file_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"d1/a"
        new_path = b"d2/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d1",), (old_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path, new_path)],
            expected_added=[(b"d2",)],
            expected_removed=[(b"d1",)],
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_rename_then_modify_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(branch, 1, expected_added=[(old_path,)])
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_rename_then_modify_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"d/a"
        new_path = b"d/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (old_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_renamed=[(old_path, new_path)]
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")

    def test_rename_then_modify_symlink_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"d1/a"
        new_path = b"d2/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d1",), (old_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path, new_path)],
            expected_added=[(b"d2",)],
            expected_removed=[(b"d1",)],
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")


class TestImportToPackDeleteRenameThenModify(TestCaseForGenericProcessor):
    """Test rename of to a deleted path then modfy the new-path in the same commit."""

    def get_command_iter(self, old_path, new_path, kind="file"):
        # Revno 1: create two files or symlinks
        # Revno 2: delete one, rename the other to it then modify the newly created path
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    old_path, kind_to_mode(kind, False), None, b"aaa"
                )
                yield commands.FileModifyCommand(
                    new_path, kind_to_mode(kind, False), None, b"zzz"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileDeleteCommand(new_path)
                yield commands.FileRenameCommand(old_path, new_path)
                yield commands.FileModifyCommand(
                    new_path, kind_to_mode(kind, False), None, b"bbb"
                )

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_delete_rename_then_modify_file_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(old_path,), (new_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"zzz")
        self.assertContent(branch, revtree2, new_path, b"bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree1, new_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_delete_rename_then_modify_file_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"d/a"
        new_path = b"d/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (old_path,), (new_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"zzz")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_delete_rename_then_modify_file_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"d1/a"
        new_path = b"d2/b"
        handler.process(self.get_command_iter(old_path, new_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d1",), (b"d2",), (old_path,), (new_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(b"d1",), (new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertContent(branch, revtree1, old_path, b"aaa")
        self.assertContent(branch, revtree1, new_path, b"zzz")
        self.assertContent(branch, revtree2, new_path, b"bbb")

    def test_delete_rename_then_modify_symlink_in_root(self):
        handler, branch = self.get_handler()
        old_path = b"a"
        new_path = b"b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(old_path,), (new_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "zzz")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")
        self.assertRevisionRoot(revtree1, old_path)
        self.assertRevisionRoot(revtree1, new_path)
        self.assertRevisionRoot(revtree2, new_path)

    def test_delete_rename_then_modify_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        old_path = b"d/a"
        new_path = b"d/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (old_path,), (new_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "zzz")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")

    def test_delete_rename_then_modify_symlink_in_new_dir(self):
        handler, branch = self.get_handler()
        old_path = b"d1/a"
        new_path = b"d2/b"
        handler.process(self.get_command_iter(old_path, new_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d1",), (b"d2",), (old_path,), (new_path,)]
        )
        # Note: the delta doesn't show the modification?
        # The actual new content is validated in the assertions following.
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_removed=[(b"d1",), (new_path,)],
            expected_renamed=[(old_path, new_path)],
        )
        self.assertSymlinkTarget(branch, revtree1, old_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, new_path, "zzz")
        self.assertSymlinkTarget(branch, revtree2, new_path, "bbb")


class TestImportToPackRenameTricky(TestCaseForGenericProcessor):
    def file_command_iter(self, path1, old_path2, new_path2, kind="file"):
        # Revno 1: create two files or symlinks in a directory
        # Revno 2: rename the second file so that it implicitly deletes the
        # first one because either:
        # * the new file is a in directory with the old file name
        # * the new file has the same name as the directory of the first
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    path1, kind_to_mode(kind, False), None, b"aaa"
                )
                yield commands.FileModifyCommand(
                    old_path2, kind_to_mode(kind, False), None, b"bbb"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileRenameCommand(old_path2, new_path2)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_rename_file_becomes_directory(self):
        handler, branch = self.get_handler()
        old_path2 = b"foo"
        path1 = b"a/b"
        new_path2 = b"a/b/c"
        handler.process(self.file_command_iter(path1, old_path2, new_path2))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path1,), (old_path2,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path2, new_path2)],
            expected_kind_changed=[(path1, "file", "directory")],
        )
        self.assertContent(branch, revtree1, path1, b"aaa")
        self.assertContent(branch, revtree2, new_path2, b"bbb")

    def test_rename_directory_becomes_file(self):
        handler, branch = self.get_handler()
        old_path2 = b"foo"
        path1 = b"a/b/c"
        new_path2 = b"a/b"
        handler.process(self.file_command_iter(path1, old_path2, new_path2))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (b"a/b",), (path1,), (old_path2,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path2, new_path2)],
            expected_removed=[(path1,), (new_path2,)],
        )
        self.assertContent(branch, revtree1, path1, b"aaa")
        self.assertContent(branch, revtree2, new_path2, b"bbb")

    def test_rename_symlink_becomes_directory(self):
        handler, branch = self.get_handler()
        old_path2 = b"foo"
        path1 = b"a/b"
        new_path2 = b"a/b/c"
        handler.process(self.file_command_iter(path1, old_path2, new_path2, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (path1,), (old_path2,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path2, new_path2)],
            expected_kind_changed=[(path1, "symlink", "directory")],
        )
        self.assertSymlinkTarget(branch, revtree1, path1, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path2, "bbb")

    def test_rename_directory_becomes_symlink(self):
        handler, branch = self.get_handler()
        old_path2 = b"foo"
        path1 = b"a/b/c"
        new_path2 = b"a/b"
        handler.process(self.file_command_iter(path1, old_path2, new_path2, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (b"a/b",), (path1,), (old_path2,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_renamed=[(old_path2, new_path2)],
            expected_removed=[(path1,), (new_path2,)],
        )
        self.assertSymlinkTarget(branch, revtree1, path1, "aaa")
        self.assertSymlinkTarget(branch, revtree2, new_path2, "bbb")


class TestImportToPackCopy(TestCaseForGenericProcessor):
    def file_command_iter(self, src_path, dest_path, kind="file"):
        # Revno 1: create a file or symlink
        # Revno 2: copy it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    src_path, kind_to_mode(kind, False), None, b"aaa"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileCopyCommand(src_path, dest_path)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_copy_file_in_root(self):
        handler, branch = self.get_handler()
        src_path = b"a"
        dest_path = b"b"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_added=[(dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree2, dest_path)

    def test_copy_file_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b"a/a"
        dest_path = b"a/b"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_added=[(dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")

    def test_copy_file_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b"a/a"
        dest_path = b"b/a"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_added=[(b"b",), (dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")

    def test_copy_symlink_in_root(self):
        handler, branch = self.get_handler()
        src_path = b"a"
        dest_path = b"b"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_added=[(dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree2, dest_path)

    def test_copy_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b"a/a"
        dest_path = b"a/b"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_added=[(dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")

    def test_copy_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b"a/a"
        dest_path = b"b/a"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_added=[(b"b",), (dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")


class TestImportToPackCopyNew(TestCaseForGenericProcessor):
    """Test copy of a newly added file."""

    def file_command_iter(self, src_path, dest_path, kind="file"):
        # Revno 1: create a file or symlink and copy it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    src_path, kind_to_mode(kind, False), None, b"aaa"
                )
                yield commands.FileCopyCommand(src_path, dest_path)

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

        return command_list

    def test_copy_new_file_in_root(self):
        handler, branch = self.get_handler()
        src_path = b"a"
        dest_path = b"b"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(src_path,), (dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree1, dest_path)

    def test_copy_new_file_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b"a/a"
        dest_path = b"a/b"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (src_path,), (dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"aaa")

    def test_copy_new_file_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b"a/a"
        dest_path = b"b/a"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (src_path,), (b"b",), (dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"aaa")

    def test_copy_new_symlink_in_root(self):
        handler, branch = self.get_handler()
        src_path = b"a"
        dest_path = b"b"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(src_path,), (dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree1, dest_path)

    def test_copy_new_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b"a/a"
        dest_path = b"a/b"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (src_path,), (dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "aaa")

    def test_copy_new_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b"a/a"
        dest_path = b"b/a"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"a",), (src_path,), (b"b",), (dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "aaa")


class TestImportToPackCopyToDeleted(TestCaseForGenericProcessor):
    def file_command_iter(self, src_path, dest_path, kind="file"):
        # Revno 1: create two files or symlinks
        # Revno 2: delete one and copy the other one to its path
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    src_path, kind_to_mode(kind, False), None, b"aaa"
                )
                yield commands.FileModifyCommand(
                    dest_path, kind_to_mode(kind, False), None, b"bbb"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileDeleteCommand(dest_path)
                yield commands.FileCopyCommand(src_path, dest_path)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_copy_to_deleted_file_in_root(self):
        handler, branch = self.get_handler()
        src_path = b"a"
        dest_path = b"b"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(src_path,), (dest_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(dest_path,)], expected_added=[(dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"bbb")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree1, dest_path)

    def test_copy_to_deleted_symlink_in_root(self):
        handler, branch = self.get_handler()
        src_path = b"a"
        dest_path = b"b"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(src_path,), (dest_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(dest_path,)], expected_added=[(dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree1, dest_path)

    def test_copy_to_deleted_file_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b"d/a"
        dest_path = b"d/b"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (src_path,), (dest_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(dest_path,)], expected_added=[(dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree1, dest_path, b"bbb")
        self.assertContent(branch, revtree2, src_path, b"aaa")
        self.assertContent(branch, revtree2, dest_path, b"aaa")

    def test_copy_to_deleted_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b"d/a"
        dest_path = b"d/b"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree0, revtree1 = self.assertChanges(
            branch, 1, expected_added=[(b"d",), (src_path,), (dest_path,)]
        )
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_removed=[(dest_path,)], expected_added=[(dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree1, dest_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "aaa")


class TestImportToPackCopyModified(TestCaseForGenericProcessor):
    """Test copy of file/symlink already modified in this commit."""

    def file_command_iter(self, src_path, dest_path, kind="file"):
        # Revno 1: create a file or symlink
        # Revno 2: modify and copy it
        def command_list():
            author = [b"", b"bugs@a.com", time.time(), time.timezone]
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    src_path, kind_to_mode(kind, False), None, b"aaa"
                )

            yield commands.CommitCommand(
                b"head", b"1", author, committer, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileModifyCommand(
                    src_path, kind_to_mode(kind, False), None, b"bbb"
                )
                yield commands.FileCopyCommand(src_path, dest_path)

            yield commands.CommitCommand(
                b"head", b"2", author, committer, b"commit 2", b":1", [], files_two
            )

        return command_list

    def test_copy_of_modified_file_in_root(self):
        handler, branch = self.get_handler()
        src_path = b"a"
        dest_path = b"b"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_modified=[(src_path,)], expected_added=[(dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"bbb")
        self.assertContent(branch, revtree2, dest_path, b"bbb")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree2, dest_path)

    def test_copy_of_modified_file_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b"d/a"
        dest_path = b"d/b"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_modified=[(src_path,)], expected_added=[(dest_path,)]
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"bbb")
        self.assertContent(branch, revtree2, dest_path, b"bbb")

    def test_copy_of_modified_file_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b"d1/a"
        dest_path = b"d2/a"
        handler.process(self.file_command_iter(src_path, dest_path))
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_modified=[(src_path,)],
            expected_added=[(b"d2",), (dest_path,)],
        )
        self.assertContent(branch, revtree1, src_path, b"aaa")
        self.assertContent(branch, revtree2, src_path, b"bbb")
        self.assertContent(branch, revtree2, dest_path, b"bbb")

    def test_copy_of_modified_symlink_in_root(self):
        handler, branch = self.get_handler()
        src_path = b"a"
        dest_path = b"b"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_modified=[(src_path,)], expected_added=[(dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "bbb")
        self.assertRevisionRoot(revtree1, src_path)
        self.assertRevisionRoot(revtree2, dest_path)

    def test_copy_of_modified_symlink_in_subdir(self):
        handler, branch = self.get_handler()
        src_path = b"d/a"
        dest_path = b"d/b"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree1, revtree2 = self.assertChanges(
            branch, 2, expected_modified=[(src_path,)], expected_added=[(dest_path,)]
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "bbb")

    def test_copy_of_modified_symlink_to_new_dir(self):
        handler, branch = self.get_handler()
        src_path = b"d1/a"
        dest_path = b"d2/a"
        handler.process(self.file_command_iter(src_path, dest_path, "symlink"))
        revtree1, revtree2 = self.assertChanges(
            branch,
            2,
            expected_modified=[(src_path,)],
            expected_added=[(b"d2",), (dest_path,)],
        )
        self.assertSymlinkTarget(branch, revtree1, src_path, "aaa")
        self.assertSymlinkTarget(branch, revtree2, src_path, "bbb")
        self.assertSymlinkTarget(branch, revtree2, dest_path, "bbb")


class TestImportToPackFileKinds(TestCaseForGenericProcessor):
    def get_command_iter(self, path, kind, content):
        def command_list():
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    path, kind_to_mode(kind, False), None, content
                )

            yield commands.CommitCommand(
                b"head", b"1", None, committer, b"commit 1", None, [], files_one
            )

        return command_list

    def test_import_plainfile(self):
        handler, branch = self.get_handler()
        handler.process(self.get_command_iter(b"foo", "file", b"aaa"))

    def test_import_symlink(self):
        handler, branch = self.get_handler()
        handler.process(self.get_command_iter(b"foo", "symlink", b"bar"))


class TestModifyRevertInBranch(TestCaseForGenericProcessor):
    def file_command_iter(self):
        # A     add 'foo'
        # |\
        # | B   modify 'foo'
        # | |
        # | C   revert 'foo' back to A
        # |/
        # D     merge 'foo'
        def command_list():
            committer_a = [b"", b"a@elmer.com", time.time(), time.timezone]
            committer_b = [b"", b"b@elmer.com", time.time(), time.timezone]
            committer_c = [b"", b"c@elmer.com", time.time(), time.timezone]
            committer_d = [b"", b"d@elmer.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    b"foo", kind_to_mode("file", False), None, b"content A\n"
                )

            yield commands.CommitCommand(
                b"head", b"1", None, committer_a, b"commit 1", None, [], files_one
            )

            def files_two():
                yield commands.FileModifyCommand(
                    b"foo", kind_to_mode("file", False), None, b"content B\n"
                )

            yield commands.CommitCommand(
                b"head", b"2", None, committer_b, b"commit 2", b":1", [], files_two
            )

            def files_three():
                yield commands.FileModifyCommand(
                    b"foo", kind_to_mode("file", False), None, b"content A\n"
                )

            yield commands.CommitCommand(
                b"head", b"3", None, committer_c, b"commit 3", b":2", [], files_three
            )
            yield commands.CommitCommand(
                b"head",
                b"4",
                None,
                committer_d,
                b"commit 4",
                b":1",
                [b":3"],
                lambda: [],
            )

        return command_list

    def test_modify_revert(self):
        handler, branch = self.get_handler()
        handler.process(self.file_command_iter())
        branch.lock_read()
        self.addCleanup(branch.unlock)
        rev_d = branch.last_revision()
        rev_a, rev_c = branch.repository.get_parent_map([rev_d])[rev_d]
        rev_b = branch.repository.get_parent_map([rev_c])[rev_c][0]
        rtree_a, rtree_b, rtree_c, rtree_d = branch.repository.revision_trees(
            [rev_a, rev_b, rev_c, rev_d]
        )
        self.assertEqual(rev_a, rtree_a.get_file_revision("foo"))
        self.assertEqual(rev_b, rtree_b.get_file_revision("foo"))
        self.assertEqual(rev_c, rtree_c.get_file_revision("foo"))
        self.assertEqual(rev_c, rtree_d.get_file_revision("foo"))


class TestCommitCommands(TestCaseForGenericProcessor):
    def test_non_utf8_commit_message(self):
        handler, branch = self.get_handler()

        def files_one():
            yield commands.FileModifyCommand(
                b"a", kind_to_mode("file", False), None, b"data"
            )

        def command_list():
            committer = [b"", b"elmer@a.com", time.time(), time.timezone]
            yield commands.CommitCommand(
                b"head",
                b"1",
                None,
                committer,
                b"This is a funky character: \x83",
                None,
                [],
                files_one,
            )

        handler.process(command_list)
        rev = branch.repository.get_revision(branch.last_revision())
        self.assertEqual("This is a funky character: \ufffd", rev.message)


class TestAddNonUtf8InBranch(TestCaseForGenericProcessor):
    def file_command_iter(self):
        # A     add 'foo\x83'
        def command_list():
            committer_a = [b"", b"a@elmer.com", time.time(), time.timezone]

            def files_one():
                yield commands.FileModifyCommand(
                    b"foo\x83", kind_to_mode("file", False), None, b"content A\n"
                )

            yield commands.CommitCommand(
                b"head", b"1", None, committer_a, b"commit 1", None, [], files_one
            )

        return command_list

    def test_add(self):
        handler, branch = self.get_handler()
        handler.process(self.file_command_iter())
        branch.lock_read()
        self.addCleanup(branch.unlock)
        rev_a = branch.last_revision()
        rtree_a = branch.repository.revision_tree(rev_a)
        self.assertEqual(rev_a, rtree_a.get_file_revision("foo\ufffd"))
